import json
import re
import wkw
import logging
import numpy as np

from argparse import ArgumentParser
from glob import iglob
from os import path, listdir
from typing import Optional


def create_parser():
    parser = ArgumentParser()

    parser.add_argument("path", help="Directory containing the dataset.")

    parser.add_argument("--name", "-n", help="Name of the dataset")

    parser.add_argument(
        "--scale",
        "-s",
        help="Scale of the dataset (e.g. 11.2,11.2,25)",
        default="1,1,1",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--compute_max_id",
        "-c",
        help="set to compute max id",
        default=False,
        action="store_true",
    )
    group.add_argument("--max_id", help="set max id of segmentation.", default=0)

    return parser


def write_webknossos_metadata(
    dataset_path,
    name,
    scale,
    max_id=0,
    compute_max_id=False,
    exact_bounding_box: Optional[dict] = None,
):

    # Generate a metadata file for webKnossos
    # Currently includes no source of information for team
    datasource_properties_path = path.join(dataset_path, "datasource-properties.json")
    layers = list(
        detect_layers(dataset_path, max_id, compute_max_id, exact_bounding_box)
    )
    with open(datasource_properties_path, "wt") as datasource_properties_json:
        json.dump(
            {
                "id": {"name": name, "team": "<unknown>"},
                "dataLayers": layers,
                "scale": scale,
            },
            datasource_properties_json,
            indent=2,
        )


def read_metadata_for_layer(wkw_path, layer_name):
    datasource_properties = json.load(
        open(path.join(wkw_path, "datasource-properties.json"), "r")
    )
    layers = datasource_properties["dataLayers"]
    layer_info = next(layer for layer in layers if layer["name"] == layer_name)
    dtype = np.dtype(layer_info["elementClass"])
    bounding_box = layer_info["boundingBox"]
    origin = bounding_box["topLeft"]
    bounding_box = [
        bounding_box["width"],
        bounding_box["height"],
        bounding_box["depth"],
    ]

    return layer_info, dtype, bounding_box, origin


def detect_dtype(dataset_path, layer, mag=1):
    layer_path = path.join(dataset_path, layer, str(mag))
    if path.exists(layer_path):
        with wkw.Dataset.open(layer_path) as dataset:
            voxel_type = dataset.header.voxel_type
            num_channels = dataset.header.num_channels
            voxel_size = np.dtype(voxel_type)
            if voxel_size == np.uint8 and num_channels > 1:
                return "uint" + str(8 * num_channels)
            else:
                return str(np.dtype(voxel_type))


def detect_cubeLength(dataset_path, layer, mag=1):
    layer_path = path.join(dataset_path, layer, str(mag))
    if path.exists(layer_path):
        with wkw.Dataset.open(layer_path) as dataset:
            return dataset.header.block_len * dataset.header.file_len


def detect_bbox(dataset_path, layer, mag=1):
    # Detect the coarse bounding box of a dataset by iterating
    # over the WKW cubes
    layer_path = path.join(dataset_path, layer, str(mag))

    def list_files(layer_path):
        return iglob(path.join(layer_path, "*", "*", "*.wkw"), recursive=True)

    def parse_cube_file_name(filename):
        CUBE_REGEX = re.compile(r"z(\d+)/y(\d+)/x(\d+)(\.wkw)$")
        m = CUBE_REGEX.search(filename)
        return (int(m.group(3)), int(m.group(2)), int(m.group(1)))

    def list_cubes(layer_path):
        return (parse_cube_file_name(f) for f in list_files(layer_path))

    xs, ys, zs = list(zip(*list_cubes(layer_path)))

    min_x, min_y, min_z = min(xs), min(ys), min(zs)
    max_x, max_y, max_z = max(xs), max(ys), max(zs)

    cubeLength = detect_cubeLength(dataset_path, layer, mag)

    return {
        "topLeft": [min_x * cubeLength, min_y * cubeLength, min_z * cubeLength],
        "width": (1 + max_x - min_x) * cubeLength,
        "height": (1 + max_y - min_y) * cubeLength,
        "depth": (1 + max_z - min_z) * cubeLength,
    }


def detect_resolutions(dataset_path, layer):
    for mag in listdir(path.join(dataset_path, layer)):
        if re.match(r"^\d+$", mag) is not None:
            yield int(mag)


def detect_standard_layer(dataset_path, layer_name, exact_bounding_box=None):
    # Perform metadata detection for well-known layers

    if exact_bounding_box is None:
        bbox = detect_bbox(dataset_path, layer_name)
    else:
        bbox = exact_bounding_box

    dtype = detect_dtype(dataset_path, layer_name)

    mags = list(detect_resolutions(dataset_path, layer_name))
    mags.sort()
    resolutions = [
        {
            "resolution": (mag,) * 3,
            "cubeLength": detect_cubeLength(dataset_path, layer_name, mag),
        }
        for mag in mags
    ]

    return {
        "dataFormat": "wkw",
        "name": layer_name,
        "category": layer_name,
        "elementClass": dtype,
        "boundingBox": bbox,
        "wkwResolutions": list(resolutions),
    }


def detect_segmentation_layer(
    dataset_path, layer_name, max_id, compute_max_id=False, exact_bounding_box=None
):
    layer_info = detect_standard_layer(dataset_path, layer_name, exact_bounding_box)
    layer_info["mappings"] = []
    layer_info["largestSegmentId"] = max_id

    if compute_max_id:
        logging.info("Computing max id of layer={}".format(layer_name))
        # Computing the current largest segment id
        # This may take very long due to IO load
        layer_path = path.join(dataset_path, layer_name, "1")
        with wkw.Dataset.open(layer_path) as dataset:
            bbox = layer_info["boundingBox"]
            layer_info["largestSegmentId"] = int(
                np.max(
                    dataset.read(
                        bbox["topLeft"], [bbox["width"], bbox["height"], bbox["depth"]]
                    )
                )
            )
            logging.info(
                "Max id of layer={} is {}".format(
                    layer_name, layer_info["largestSegmentId"]
                )
            )

    return layer_info


def detect_layers(dataset_path, max_id, compute_max_id, exact_bounding_box=None):
    # Detect metadata for well-known layers, e.g. color and segmentation
    if path.exists(path.join(dataset_path, "color")):
        yield detect_standard_layer(dataset_path, "color", exact_bounding_box)
    if path.exists(path.join(dataset_path, "segmentation")):
        yield detect_segmentation_layer(
            dataset_path, "segmentation", max_id, compute_max_id, exact_bounding_box
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    args = create_parser().parse_args()
    scale = tuple(float(x) for x in args.scale.split(","))
    write_webknossos_metadata(
        args.path, args.name, scale, args.max_id, args.compute_max_id
    )



import numpy as np
from wkcuber.utils import get_chunks, get_regular_chunks

BLOCK_LEN = 32


def test_get_chunks():
    source = list(range(0, 48))
    target = list(get_chunks(source, 8))

    assert len(target) == 6
    assert target[0] == list(range(0, 8))


def test_get_regular_chunks():
    target = list(get_regular_chunks(4, 44, 8))

    assert len(target) == 6
    assert list(target[0]) == list(range(0, 8))
    assert list(target[-1]) == list(range(40, 48))



import logging
import numpy as np
from wkcuber.downsampling import (
    InterpolationModes,
    downsample_cube,
    downsample_cube_job,
    cube_addresses,
)
import wkw
from wkcuber.utils import WkwDatasetInfo, open_wkw
import shutil

WKW_CUBE_SIZE = 1024
CUBE_EDGE_LEN = 256

source_info = WkwDatasetInfo("testdata/WT1_wkw", "color", "uint8", 1)
target_info = WkwDatasetInfo("testoutput/WT1_wkw", "color", "uint8", 2)


def read_wkw(wkw_info, offset, size, **kwargs):
    with open_wkw(wkw_info, **kwargs) as wkw_dataset:
        return wkw_dataset.read(offset, size)


def test_downsample_cube():
    buffer = np.zeros((CUBE_EDGE_LEN,) * 3, dtype=np.uint8)
    buffer[:, :, :] = np.arange(0, CUBE_EDGE_LEN)

    output = downsample_cube(buffer, 2, InterpolationModes.MEDIAN)

    assert output.shape == (CUBE_EDGE_LEN // 2,) * 3
    assert buffer[0, 0, 0] == 0
    assert buffer[0, 0, 1] == 1
    assert np.all(output[:, :, :] == np.arange(0, CUBE_EDGE_LEN, 2))


def test_cube_addresses():
    addresses = cube_addresses(source_info)
    assert len(addresses) == 5 * 5 * 1

    assert min(addresses) == (0, 0, 0)
    assert max(addresses) == (4, 4, 0)


def downsample_test_helper(use_compress):
    try:
        shutil.rmtree(target_info.dataset_path)
    except:
        pass

    offset = (1, 2, 0)
    source_buffer = read_wkw(
        source_info,
        tuple(a * WKW_CUBE_SIZE * 2 for a in offset),
        (CUBE_EDGE_LEN * 2,) * 3,
    )[0]
    assert np.any(source_buffer != 0)

    downsample_cube_job(
        source_info,
        target_info,
        2,
        InterpolationModes.MAX,
        CUBE_EDGE_LEN,
        offset,
        use_compress,
    )

    assert np.any(source_buffer != 0)
    block_type = (
        wkw.Header.BLOCK_TYPE_LZ4HC if use_compress else wkw.Header.BLOCK_TYPE_RAW
    )
    target_buffer = read_wkw(
        target_info,
        tuple(a * WKW_CUBE_SIZE for a in offset),
        (CUBE_EDGE_LEN,) * 3,
        block_type=block_type,
    )[0]
    assert np.any(target_buffer != 0)

    assert np.all(
        target_buffer == downsample_cube(source_buffer, 2, InterpolationModes.MAX)
    )


def test_downsample_cube_job():
    downsample_test_helper(False)


def test_compressed_downsample_cube_job():
    downsample_test_helper(True)


def test_downsample_multi_channel():
    source_info = WkwDatasetInfo("testoutput/multi-channel-test", "color", "uint8", 1)
    target_info = WkwDatasetInfo("testoutput/multi-channel-test", "color", "uint8", 2)
    try:
        shutil.rmtree(source_info.dataset_path)
        shutil.rmtree(target_info.dataset_path)
    except:
        pass

    offset = (0, 0, 0)
    num_channels = 3
    size = (32, 32, 10)
    source_data = (128 * np.random.randn(num_channels, size[0], size[1], size[2])).astype('uint8')
    file_len = 32

    with open_wkw(source_info, num_channels=num_channels, file_len=file_len) as wkw_dataset:
        print("writing source_data shape", source_data.shape)
        wkw_dataset.write(offset, source_data)
    assert np.any(source_data != 0)

    downsample_cube_job(
        source_info,
        target_info,
        2,
        InterpolationModes.MAX,
        CUBE_EDGE_LEN,
        tuple(a * WKW_CUBE_SIZE for a in offset),
        False,
    )

    channels = []
    for channel_index in range(num_channels):
        channels.append(downsample_cube(source_data[channel_index], 2, InterpolationModes.MAX))
    joined_buffer = np.stack(channels)

    target_buffer = read_wkw(
        target_info,
        tuple(a * WKW_CUBE_SIZE for a in offset),
        list(map(lambda x: x // 2, size)),
        file_len=file_len
    )
    assert np.any(target_buffer != 0)

    assert np.all(
        target_buffer == joined_buffer
    )



from argparse import ArgumentParser
from os import path
from uuid import uuid4
import logging
import shutil

from .cubing import cubing, BLOCK_LEN
from .downsampling import downsample_mags, DEFAULT_EDGE_LEN
from .compress import compress_mag
from .metadata import write_webknossos_metadata
from .utils import add_verbose_flag, add_jobs_flag


def create_parser():
    parser = ArgumentParser()

    parser.add_argument("source_path", help="Directory containing the input images.")

    parser.add_argument(
        "target_path", help="Output directory for the generated dataset."
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--dtype",
        "-d",
        help="Target datatype (e.g. uint8, uint16, uint32)",
        default="uint8",
    )

    parser.add_argument(
        "--batch_size",
        "-b",
        type=int,
        help="Number of slices to buffer per job",
        default=BLOCK_LEN,
    )

    parser.add_argument(
        "--max_mag",
        "-m",
        help="Max resolution to be downsampled. Needs to be a power of 2.",
        type=int,
        default=512,
    )

    parser.add_argument(
        "--no_compress",
        help="Don't compress this data",
        default=False,
        action="store_true",
    )

    parser.add_argument("--name", "-n", help="Name of the dataset", default=None)

    parser.add_argument(
        "--scale",
        "-s",
        help="Scale of the dataset (e.g. 11.2,11.2,25). This is the size of one voxel in nm.",
        default="1,1,1",
    )

    add_verbose_flag(parser)
    add_jobs_flag(parser)

    return parser


def compress_mag_inplace(target_path, layer_name, mag, jobs):
    compress_target_path = "{}.compress-{}".format(target_path, uuid4())
    compress_mag(target_path, layer_name, compress_target_path, mag, jobs)

    shutil.rmtree(path.join(args.target_path, args.layer_name, str(mag)))
    shutil.move(
        path.join(compress_target_path, layer_name, str(mag)),
        path.join(target_path, layer_name, str(mag)),
    )
    shutil.rmtree(compress_target_path)


if __name__ == "__main__":
    args = create_parser().parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    bounding_box = cubing(
        args.source_path,
        args.target_path,
        args.layer_name,
        args.dtype,
        args.batch_size,
        args.jobs,
    )

    if not args.no_compress:
        compress_mag_inplace(args.target_path, args.layer_name, 1, args.jobs)

    downsample_mags(
        args.target_path,
        args.layer_name,
        1,
        args.max_mag,
        args.dtype,
        "default",
        DEFAULT_EDGE_LEN,
        args.jobs,
        not args.no_compress,
    )

    scale = tuple(float(x) for x in args.scale.split(","))
    write_webknossos_metadata(
        args.target_path,
        args.name,
        scale,
        compute_max_id=False,
        exact_bounding_box=bounding_box,
    )



import time
import logging
import numpy as np
import re
import wkw
from argparse import ArgumentParser
from os import path, listdir
from PIL import Image

from .utils import (
    get_chunks,
    get_regular_chunks,
    find_files,
    add_verbose_flag,
    add_jobs_flag,
    pool_get_lock,
    open_wkw,
    WkwDatasetInfo,
    ParallelExecutor,
)
from .cubing import create_parser, read_image_file
from .image_readers import image_reader

BLOCK_LEN = 32
CUBE_REGEX = re.compile("(\d+)/(\d+)\.([a-zA-Z]{3,4})$")


def find_source_sections(source_path):
    section_folders = [
        f for f in listdir(source_path) if path.isdir(path.join(source_path, f))
    ]
    section_folders = [path.join(source_path, s) for s in section_folders]
    section_folders.sort()
    return section_folders


def parse_tile_file_name(filename):
    m = CUBE_REGEX.search(filename)
    if m is None:
        return None
    return (int(m.group(1)), int(m.group(2)), m.group(3))


def find_source_files(source_section_path):
    all_source_files = [
        path.join(source_section_path, s)
        for s in find_files(
            path.join(source_section_path, "**", "*"), image_reader.readers.keys()
        )
    ]

    all_source_files.sort()
    return all_source_files


def tile_cubing_job(target_wkw_info, z_batches, source_path, batch_size, tile_size):
    if len(z_batches) == 0:
        return

    with open_wkw(target_wkw_info, pool_get_lock()) as target_wkw:
        # Iterate over the z batches
        # Batching is useful to utilize IO more efficiently
        for z_batch in get_chunks(z_batches, batch_size):
            try:
                ref_time = time.time()
                logging.info("Cubing z={}-{}".format(z_batch[0], z_batch[-1]))

                # Detect all available tiles in this z batch
                tile_coords = []
                for z in z_batch:
                    tile_coords += [
                        (z,) + parse_tile_file_name(f)
                        for f in find_source_files("{}/{}".format(source_path, z))
                    ]
                # Figure out what x-y combinations are available in this z batch
                xy = sorted(set([(x, y) for z, y, x, _ in tile_coords]))

                # Iterate over all x-y combinations from this z batch
                for x, y in xy:
                    ref_time2 = time.time()
                    buffer = []
                    for z in z_batch:
                        # Find the file extension of the x-y tile in this z
                        ext = next(
                            (
                                ext
                                for _z, _y, _x, ext in tile_coords
                                if _z == z and _y == y and _x == x
                            ),
                            None,
                        )
                        # Read file if exists or zeros instead
                        if ext is not None:
                            image = read_image_file(
                                "{}/{}/{}/{}.{}".format(source_path, z, y, x, ext),
                                target_wkw_info.dtype,
                            )
                            buffer.append(image)
                        else:
                            buffer.append(
                                np.zeros(tile_size, dtype=target_wkw_info.dtype)
                            )

                    # Write buffer to target
                    buffer = np.dstack(buffer)
                    if np.any(buffer != 0):
                        target_wkw.write(
                            [x * tile_size[0], y * tile_size[1], z_batch[0]], buffer
                        )
                    logging.debug(
                        "Cubing of z={}-{} x={} y={} took {:.8f}s".format(
                            z_batch[0], z_batch[-1], x, y, time.time() - ref_time2
                        )
                    )

                logging.debug(
                    "Cubing of z={}-{} took {:.8f}s".format(
                        z_batch[0], z_batch[-1], time.time() - ref_time
                    )
                )

            except Exception as exc:
                logging.error(
                    "Cubing of z={}-{} failed with {}".format(
                        z_batch[0], z_batch[-1], exc
                    )
                )
                raise exc


def tile_cubing(source_path, target_path, layer_name, dtype, batch_size, jobs):
    # Detect available z sections
    sections = find_source_sections(source_path)
    if len(sections) == 0:
        logging.error("No source files found")
        return
    min_z = min([int(path.basename(f)) for f in sections])
    max_z = max([int(path.basename(f)) for f in sections])

    # Determine tile size from first matching file
    tile_size = image_reader.read_dimensions(
        next(find_files(path.join(source_path, "**", "*"), image_reader.readers.keys()))
    )
    logging.info(
        "Found source files: count={} tile_size={}x{}".format(
            len(sections), tile_size[0], tile_size[1]
        )
    )

    target_wkw_info = WkwDatasetInfo(target_path, layer_name, dtype, 1)
    with ParallelExecutor(jobs) as pool:
        # Iterate over all z batches
        for z_batch in get_regular_chunks(min_z, max_z, BLOCK_LEN):
            pool.submit(
                tile_cubing_job,
                target_wkw_info,
                list(z_batch),
                source_path,
                int(batch_size),
                tile_size,
            )


if __name__ == "__main__":
    args = create_parser().parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    tile_cubing(
        args.source_path,
        args.target_path,
        args.layer_name,
        args.dtype,
        int(args.batch_size),
        int(args.jobs),
    )



import logging
import wkw
import re
import numpy as np
from argparse import ArgumentParser
from math import floor, log2
from os import path, listdir
from scipy.stats import mode
from scipy.ndimage.interpolation import zoom
from itertools import product
from functools import lru_cache
from enum import Enum

from .utils import (
    add_jobs_flag,
    add_verbose_flag,
    open_wkw,
    WkwDatasetInfo,
    ParallelExecutor,
    pool_get_lock,
    time_start,
    time_stop,
)

DEFAULT_EDGE_LEN = 256
CUBE_REGEX = re.compile(r"z(\d+)/y(\d+)/x(\d+)(\.wkw)$")


def parse_cube_file_name(filename):
    m = CUBE_REGEX.search(filename)
    return (int(m.group(3)), int(m.group(2)), int(m.group(1)))


class InterpolationModes(Enum):
    MEDIAN = 0
    MODE = 1
    NEAREST = 2
    BILINEAR = 3
    BICUBIC = 4
    MAX = 5
    MIN = 6


def create_parser():
    parser = ArgumentParser()

    parser.add_argument("path", help="Directory containing the dataset.")

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--interpolation_mode",
        "-i",
        help="Interpolation mode (median, mode, nearest, bilinear or bicubic)",
        default="default",
    )

    parser.add_argument(
        "--dtype",
        "-d",
        help="Target datatype (e.g. uint8, uint16, uint32)",
        default="uint8",
    )

    parser.add_argument(
        "--max", "-m", help="Max resolution to be downsampled", type=int, default=512
    )

    parser.add_argument(
        "--from_mag",
        "--from",
        "-f",
        help="Resolution to base downsampling on",
        type=int,
        default=1,
    )

    parser.add_argument(
        "--buffer_cube_size",
        "-b",
        help="Size of buffered cube to be downsampled (i.e. buffer cube edge length)",
        type=int,
        default=DEFAULT_EDGE_LEN,
    )

    parser.add_argument(
        "--compress", action="store_true", help="Compress data during downsampling"
    )

    add_jobs_flag(parser)
    add_verbose_flag(parser)

    return parser


def cube_addresses(source_wkw_info):
    # Gathers all WKW cubes in the dataset
    with open_wkw(source_wkw_info) as source_wkw:
        wkw_addresses = list(parse_cube_file_name(f) for f in source_wkw.list_files())
        wkw_addresses.sort()
        return wkw_addresses


def downsample(
    source_wkw_info,
    target_wkw_info,
    source_mag,
    target_mag,
    interpolation_mode,
    cube_edge_len,
    jobs,
    compress,
):
    assert source_mag < target_mag
    logging.info("Downsampling mag {} from mag {}".format(target_mag, source_mag))

    mag_factor = int(target_mag / source_mag)
    # Detect the cubes that we want to downsample
    source_cube_addresses = cube_addresses(source_wkw_info)

    target_cube_addresses = list(
        set(tuple(x // mag_factor for x in xyz) for xyz in source_cube_addresses)
    )
    target_cube_addresses.sort()
    logging.debug(
        "Found source cubes: count={} size={} min={} max={}".format(
            len(source_cube_addresses),
            (cube_edge_len,) * 3,
            min(source_cube_addresses),
            max(source_cube_addresses),
        )
    )
    logging.debug(
        "Found target cubes: count={} size={} min={} max={}".format(
            len(target_cube_addresses),
            (cube_edge_len,) * 3,
            min(target_cube_addresses),
            max(target_cube_addresses),
        )
    )

    with ParallelExecutor(jobs) as pool:
        for target_cube_xyz in target_cube_addresses:
            pool.submit(
                downsample_cube_job,
                source_wkw_info,
                target_wkw_info,
                mag_factor,
                interpolation_mode,
                cube_edge_len,
                target_cube_xyz,
                compress,
            )

    logging.info("Mag {0} succesfully cubed".format(target_mag))


def downsample_cube_job(
    source_wkw_info,
    target_wkw_info,
    mag_factor,
    interpolation_mode,
    cube_edge_len,
    target_cube_xyz,
    compress,
):
    logging.info("Downsampling of {}".format(target_cube_xyz))

    try:
        time_start("Downsampling of {}".format(target_cube_xyz))
        header_block_type = (
            wkw.Header.BLOCK_TYPE_LZ4HC if compress else wkw.Header.BLOCK_TYPE_RAW
        )

        with open_wkw(source_wkw_info) as source_wkw:
            num_channels = source_wkw.header.num_channels
            with open_wkw(
                target_wkw_info,
                pool_get_lock(),
                block_type=header_block_type,
                num_channels=num_channels,
            ) as target_wkw:
                wkw_cubelength = (
                    source_wkw.header.file_len * source_wkw.header.block_len
                )
                shape = (num_channels,) + (wkw_cubelength,) * 3
                file_buffer = np.zeros(shape, target_wkw_info.dtype)
                tile_length = cube_edge_len
                tile_count_per_dim = wkw_cubelength // tile_length
                assert (
                    wkw_cubelength % cube_edge_len == 0
                ), "buffer_cube_size must be a divisor of wkw cube length"

                tile_indices = list(range(0, tile_count_per_dim))
                tiles = product(tile_indices, tile_indices, tile_indices)
                file_offset = wkw_cubelength * np.array(target_cube_xyz)

                for tile in tiles:
                    target_offset = np.array(
                        tile
                    ) * tile_length + wkw_cubelength * np.array(target_cube_xyz)
                    source_offset = mag_factor * target_offset

                    # Read source buffer
                    cube_buffer_channels = source_wkw.read(
                        source_offset,
                        (wkw_cubelength * mag_factor // tile_count_per_dim,) * 3,
                    )

                    for channel_index in range(num_channels):
                        cube_buffer = cube_buffer_channels[channel_index]

                        if np.all(cube_buffer == 0):
                            logging.debug(
                                "        Skipping empty cube {} (tile {})".format(
                                    target_cube_xyz, tile
                                )
                            )
                        else:
                            # Downsample the buffer

                            data_cube = downsample_cube(
                                cube_buffer, mag_factor, interpolation_mode
                            )

                            buffer_offset = target_offset - file_offset
                            buffer_end = buffer_offset + tile_length

                            file_buffer[
                                channel_index,
                                buffer_offset[0] : buffer_end[0],
                                buffer_offset[1] : buffer_end[1],
                                buffer_offset[2] : buffer_end[2],
                            ] = data_cube

                # Write the downsampled buffer to target
                target_wkw.write(file_offset, file_buffer)
        time_stop("Downsampling of {}".format(target_cube_xyz))

    except Exception as exc:
        logging.error("Downsampling of {} failed with {}".format(target_cube_xyz, exc))
        raise exc


def non_linear_filter_3d(data, factor, func):
    ds = data.shape
    assert not any((d % factor > 0 for d in ds))
    data = data.reshape((ds[0], factor, ds[1] // factor, ds[2]), order="F")
    data = data.swapaxes(0, 1)
    data = data.reshape(
        (factor * factor, ds[0] * ds[1] // (factor * factor), factor, ds[2] // factor),
        order="F",
    )
    data = data.swapaxes(2, 1)
    data = data.reshape(
        (
            factor * factor * factor,
            (ds[0] * ds[1] * ds[2]) // (factor * factor * factor),
        ),
        order="F",
    )
    data = func(data)
    data = data.reshape((ds[0] // factor, ds[1] // factor, ds[2] // factor), order="F")
    return data


def linear_filter_3d(data, factor, order):
    ds = data.shape
    assert not any((d % factor > 0 for d in ds))
    return zoom(
        data,
        1 / factor,
        output=data.dtype,
        # 0: nearest
        # 1: bilinear
        # 2: bicubic
        order=order,
        # this does not mean nearest interpolation,
        # it corresponds to how the borders are treated.
        mode="nearest",
        prefilter=True,
    )


def _max(x):
    return np.max(x, axis=0)


def _min(x):
    return np.min(x, axis=0)


def _median(x):
    return np.median(x, axis=0).astype(x.dtype)


def _mode(x):
    return mode(x, axis=0, nan_policy="omit")[0][0]


def downsample_cube(cube_buffer, factor, interpolation_mode):
    if interpolation_mode == InterpolationModes.MODE:
        return non_linear_filter_3d(cube_buffer, factor, _mode)
    elif interpolation_mode == InterpolationModes.MEDIAN:
        return non_linear_filter_3d(cube_buffer, factor, _median)
    elif interpolation_mode == InterpolationModes.NEAREST:
        return linear_filter_3d(cube_buffer, factor, 0)
    elif interpolation_mode == InterpolationModes.BILINEAR:
        return linear_filter_3d(cube_buffer, factor, 1)
    elif interpolation_mode == InterpolationModes.BICUBIC:
        return linear_filter_3d(cube_buffer, factor, 2)
    elif interpolation_mode == InterpolationModes.MAX:
        return non_linear_filter_3d(cube_buffer, factor, _max)
    elif interpolation_mode == InterpolationModes.MIN:
        return non_linear_filter_3d(cube_buffer, factor, _min)
    else:
        raise Exception("Invalid interpolation mode: {}".format(interpolation_mode))


def downsample_mag(
    path,
    layer_name,
    source_mag,
    target_mag,
    dtype="uint8",
    interpolation_mode="default",
    cube_edge_len=DEFAULT_EDGE_LEN,
    jobs=1,
    compress=False,
):
    if interpolation_mode == "default":
        interpolation_mode = (
            InterpolationModes.MEDIAN
            if layer_name == "color"
            else InterpolationModes.MODE
        )
    else:
        interpolation_mode = InterpolationModes[interpolation_mode.upper()]

    source_wkw_info = WkwDatasetInfo(path, layer_name, dtype, source_mag)
    target_wkw_info = WkwDatasetInfo(path, layer_name, dtype, target_mag)
    downsample(
        source_wkw_info,
        target_wkw_info,
        source_mag,
        target_mag,
        interpolation_mode,
        cube_edge_len,
        jobs,
        compress,
    )


def downsample_mags(
    path,
    layer_name,
    from_mag,
    max_mag,
    dtype,
    interpolation_mode,
    cube_edge_len,
    jobs,
    compress,
):
    assert log2(from_mag) % 1 == 0, "'from_mag' needs to be power of 2."
    target_mag = from_mag * 2
    while target_mag <= max_mag:
        source_mag = target_mag // 2
        downsample_mag(
            path,
            layer_name,
            source_mag,
            target_mag,
            dtype,
            interpolation_mode,
            cube_edge_len,
            jobs,
            compress,
        )
        target_mag = target_mag * 2


if __name__ == "__main__":
    args = create_parser().parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    downsample_mags(
        args.path,
        args.layer_name,
        args.from_mag,
        args.max,
        args.dtype,
        args.interpolation_mode,
        args.buffer_cube_size,
        args.jobs,
        args.compress,
    )



import time
import wkw
import numpy as np
import logging
from glob import iglob
from collections import namedtuple
from multiprocessing import cpu_count, Lock
from concurrent.futures import ProcessPoolExecutor
from os import path
from platform import python_version
from math import floor, ceil


from .knossos import KnossosDataset, CUBE_EDGE_LEN


WkwDatasetInfo = namedtuple(
    "WkwDatasetInfo", ("dataset_path", "layer_name", "dtype", "mag")
)
KnossosDatasetInfo = namedtuple("KnossosDatasetInfo", ("dataset_path", "dtype"))


def _open_wkw(info, **kwargs):
    header = wkw.Header(np.dtype(info.dtype), **kwargs)
    ds = wkw.Dataset.open(
        path.join(info.dataset_path, info.layer_name, str(info.mag)), header
    )
    return ds


def open_wkw(info, lock=None, **kwargs):
    if lock is None:
        # Create dummy lock
        lock = Lock()

    with lock:
        return _open_wkw(info, **kwargs)


def open_knossos(info):
    return KnossosDataset.open(info.dataset_path, np.dtype(info.dtype))


def add_verbose_flag(parser):
    parser.add_argument(
        "--verbose", "-v", help="Verbose output", dest="verbose", action="store_true"
    )

    parser.set_defaults(verbose=False)


def find_files(source_path, extensions):
    # Find all files with a matching file extension
    return (
        f
        for f in iglob(source_path, recursive=True)
        if any([f.endswith(suffix) for suffix in extensions])
    )


def get_chunks(arr, chunk_size):
    for i in range(0, len(arr), chunk_size):
        yield arr[i : i + chunk_size]


def get_regular_chunks(min_z, max_z, chunk_size):
    i = floor(min_z / chunk_size) * chunk_size
    while i < ceil(max_z / chunk_size) * chunk_size:
        yield range(i, i + chunk_size)
        i += chunk_size


def add_jobs_flag(parser):
    parser.add_argument(
        "--jobs", "-j", help="Parallel jobs", type=int, default=cpu_count()
    )


def pool_init(lock):
    global process_pool_lock
    process_pool_lock = lock


def pool_get_lock():
    global process_pool_lock
    try:
        return process_pool_lock
    except NameError:
        return None


class ParallelExecutor:
    def __init__(self, jobs):
        self.lock = Lock()
        if python_version() >= "3.7.0":
            self.exec = ProcessPoolExecutor(
                jobs, initializer=pool_init, initargs=(self.lock,)
            )
        else:
            self.exec = ProcessPoolExecutor(jobs)
        self.futures = []

    def submit(self, fn, *args):
        future = self.exec.submit(fn, *args)
        self.futures.append(future)
        return future

    def __enter__(self):
        self.exec.__enter__()
        return self

    def __exit__(self, type, value, tb):
        [f.result() for f in self.futures]
        self.exec.__exit__(type, value, tb)


times = {}


def time_start(identifier):
    times[identifier] = time.time()


def time_stop(identifier):
    _time = times.pop(identifier)
    logging.debug("{} took {:.8f}s".format(identifier, time.time() - _time))



import time
import logging
import wkw
import numpy as np
from argparse import ArgumentParser
from os import path

from .utils import (
    add_jobs_flag,
    add_verbose_flag,
    open_wkw,
    open_knossos,
    WkwDatasetInfo,
    KnossosDatasetInfo,
    ParallelExecutor,
    pool_get_lock,
)
from .knossos import KnossosDataset, CUBE_EDGE_LEN


def create_parser():
    parser = ArgumentParser()

    parser.add_argument(
        "source_path", help="Directory containing the source KNOSSOS dataset."
    )

    parser.add_argument(
        "target_path", help="Output directory for the generated WKW dataset."
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--dtype",
        "-d",
        help="Target datatype (e.g. uint8, uint16, uint32)",
        default="uint8",
    )

    parser.add_argument("--mag", "-m", help="Magnification level", type=int, default=1)

    add_jobs_flag(parser)
    add_verbose_flag(parser)

    return parser


def convert_cube_job(cube_xyz, source_knossos_info, target_wkw_info):
    logging.info("Converting {},{},{}".format(cube_xyz[0], cube_xyz[1], cube_xyz[2]))
    ref_time = time.time()
    offset = tuple(x * CUBE_EDGE_LEN for x in cube_xyz)
    size = (CUBE_EDGE_LEN,) * 3

    with open_knossos(source_knossos_info) as source_knossos, open_wkw(
        target_wkw_info, pool_get_lock()
    ) as target_wkw:
        cube_data = source_knossos.read(offset, size)
        target_wkw.write(offset, cube_data)
    logging.debug(
        "Converting of {},{},{} took {:.8f}s".format(
            cube_xyz[0], cube_xyz[1], cube_xyz[2], time.time() - ref_time
        )
    )


def convert_knossos(source_path, target_path, layer_name, dtype, mag=1, jobs=1):
    source_knossos_info = KnossosDatasetInfo(source_path, dtype)
    target_wkw_info = WkwDatasetInfo(target_path, layer_name, dtype, mag)

    with open_knossos(source_knossos_info) as source_knossos, ParallelExecutor(
        jobs
    ) as pool:
        knossos_cubes = list(source_knossos.list_cubes())
        if len(knossos_cubes) == 0:
            logging.error("No input KNOSSOS cubes found.")
            exit(1)

        knossos_cubes.sort()
        for cube_xyz in knossos_cubes:
            pool.submit(
                convert_cube_job, cube_xyz, source_knossos_info, target_wkw_info
            )


if __name__ == "__main__":
    args = create_parser().parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    convert_knossos(
        args.source_path,
        args.target_path,
        args.layer_name,
        args.dtype,
        args.mag,
        args.jobs,
    )



from setuptools import setup, find_packages

setup(
    name='wkcuber',
    packages=find_packages(exclude=("tests",)),
    version='0.1.6',
    install_requires=[
        'scipy',
        'numpy',
        'pillow',
        'pyyaml',
        'wkw'
    ],
    description='A cubing tool for webKnossos',
    author='Norman Rzepka',
    author_email='norman.rzepka@scalableminds.com',
    url='https://scalableminds.com'
)



import time
import logging
import numpy as np
from argparse import ArgumentParser
from os import path

from .utils import (
    get_chunks,
    find_files,
    add_verbose_flag,
    add_jobs_flag,
    open_wkw,
    WkwDatasetInfo,
    ParallelExecutor,
    pool_get_lock,
)
from .image_readers import image_reader

BLOCK_LEN = 32


def create_parser():
    parser = ArgumentParser()

    parser.add_argument("source_path", help="Directory containing the input images.")

    parser.add_argument(
        "target_path", help="Output directory for the generated dataset."
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--dtype",
        "-d",
        help="Target datatype (e.g. uint8, uint16, uint32)",
        default="uint8",
    )

    parser.add_argument(
        "--batch_size",
        "-b",
        help="Number of sections to buffer per job",
        type=int,
        default=BLOCK_LEN,
    )

    add_verbose_flag(parser)
    add_jobs_flag(parser)

    return parser


def find_source_filenames(source_path):
    # Find all files in a folder that have a matching file extension
    source_files = list(
        find_files(path.join(source_path, "*"), image_reader.readers.keys())
    )
    assert (
        len(source_files) > 0
    ), f"No image files found in path {source_path}. Supported suffixes are {str(image_reader.readers.keys())}."
    source_files.sort()
    return source_files


def read_image_file(file_name, dtype):
    try:
        return image_reader.read_array(file_name, dtype)
    except Exception as exc:
        logging.error("Reading of file={} failed with {}".format(file_name, exc))
        raise exc


def cubing_job(
    target_wkw_info,
    z_batches,
    source_file_batches,
    batch_size,
    image_size,
    num_channels,
):
    if len(z_batches) == 0:
        return

    with open_wkw(
        target_wkw_info, pool_get_lock(), num_channels=num_channels
    ) as target_wkw:
        # Iterate over batches of continuous z sections
        # The batches have a maximum size of `batch_size`
        # Batched iterations allows to utilize IO more efficiently
        for z_batch, source_file_batch in zip(
            get_chunks(z_batches, batch_size),
            get_chunks(source_file_batches, batch_size),
        ):
            try:
                ref_time = time.time()
                logging.info("Cubing z={}-{}".format(z_batch[0], z_batch[-1]))
                buffer = []
                # Iterate over each z section in the batch
                for z, file_name in zip(z_batch, source_file_batch):
                    # Image shape will be (x, y, channel_count, z=1) or (x, y, z=1)
                    image = read_image_file(file_name, target_wkw_info.dtype)
                    assert (
                        image.shape[0:2] == image_size
                    ), "Section z={} has the wrong dimensions: {} (expected {}).".format(
                        z, image.shape, image_size
                    )
                    buffer.append(image)

                # Write batch buffer which will have shape (x, y, channel_count, z)
                # since we concat along the last axis (z)
                buffer = np.concatenate(buffer, axis=-1)

                if buffer.ndim == 4:
                    # In case of multi-channel data, we transpose the data
                    # so that the first dimension is the channel, since the wkw
                    # lib expects this.
                    # New shape will be (channel_count, x, y, z)
                    buffer = np.transpose(buffer, (2, 0, 1, 3))
                    assert buffer.shape[0] == num_channels

                target_wkw.write([0, 0, z_batch[0]], buffer)
                logging.debug(
                    "Cubing of z={}-{} took {:.8f}s".format(
                        z_batch[0], z_batch[-1], time.time() - ref_time
                    )
                )

            except Exception as exc:
                logging.error(
                    "Cubing of z={}-{} failed with {}".format(
                        z_batch[0], z_batch[-1], exc
                    )
                )
                raise exc


def cubing(source_path, target_path, layer_name, dtype, batch_size, jobs) -> dict:

    target_wkw_info = WkwDatasetInfo(target_path, layer_name, dtype, 1)
    source_files = find_source_filenames(source_path)

    # All images are assumed to have equal dimensions
    num_x, num_y = image_reader.read_dimensions(source_files[0])
    num_channels = image_reader.read_channel_count(source_files[0])
    num_z = len(source_files)

    logging.info("Found source files: count={} size={}x{}".format(num_z, num_x, num_y))
    with ParallelExecutor(jobs) as pool:
        # We iterate over all z sections
        for z in range(0, num_z, BLOCK_LEN):
            # Prepare z batches
            max_z = min(num_z, z + BLOCK_LEN)
            z_batch = list(range(z, max_z))
            # Execute
            pool.submit(
                cubing_job,
                target_wkw_info,
                z_batch,
                source_files[z:max_z],
                batch_size,
                (num_x, num_y),
                num_channels,
            )

    # Return Bounding Box
    return {"topLeft": [0, 0, 0], "width": num_x, "height": num_y, "depth": num_z}


if __name__ == "__main__":
    args = create_parser().parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    cubing(
        args.source_path,
        args.target_path,
        args.layer_name,
        args.dtype,
        args.batch_size,
        args.jobs,
    )
